import logging
import re
import time

from concurrent.futures import ThreadPoolExecutor
import functools
from lxml.etree import ElementTree
from typing import List, Dict, Pattern

from jnpr.junos import Device
from ruxit.api.base_plugin import RemoteBasePlugin
from ruxit.api.exceptions import ConfigException
from ruxit.api.topology_builder import Group, Device as Node

# Ignore warnings from the paramiko library, used by jjpr.junos
import warnings

warnings.filterwarnings(action="ignore", module=".*paramiko.*")


logger = logging.getLogger(__name__)

COMMANDS_TIMEOUT = 50


def exception_logger(function):
    @functools.wraps(function)
    def wrapper(*args, **kwargs):
        try:
            return function(*args, **kwargs)
        except Exception as e:
            logger.exception(f'Error executing {function.__name__}: "{e}"')
            raise

    return wrapper


class Measurement:
    def __init__(self, value: float):
        self.value = value
        self.timestamp = time.time()

    def __repr__(self):
        return f"Measurement({self.timestamp}, {self.value})"


class CustomJuniperPluginRemote(RemoteBasePlugin):
    """
    The Juniper Plugin will gather facts and metrics for Juniper Devices
    Currently it uses the RPC version of the commands
    The current commands are implemented:
    - dev.facts - Model, Version, Hostname, etc
    - show system connections - Gather ports
    - show system processes brief - Gather CPU Load
    - show system statistics local - Get general traffic statistics
    - show interfaces extensive - Traffic Statistics for Physical and Logical interfaces
    - show system memory local - Get Memory Usage information

    Devices model tested: vQFX, vSRX
    """

    config: Dict = None
    device: Device = None
    group: Group = None
    ips = set()
    node_name: str = None
    node: Node = None
    regex_ignore: List[Pattern] = None
    regex_include: List[Pattern] = None
    start_time: float = None

    metrics_to_collect = {
        # System Memory
        "system-memory-wired-percent": {"result_type": "absolute", "key": "mem-wired"},
        "system-memory-active-percent": {"result_type": "absolute", "key": "mem-active"},
        "load-average-one": {"result_type": "absolute", "key": "abc"},
        # TCP/UDP statistics
        "packets-sent": {"result_type": "relative", "key": "packets-sent"},
        "packets-received": {"result_type": "relative", "key": "packets-received"},
        "connection-requests": {"result_type": "relative", "key": "connection-requests"},
        "connection-accepts": {"result_type": "relative", "key": "connection-accepts"},
        "connections-established": {"result_type": "relative", "key": "connections-established"},
        "connections-closed": {"result_type": "relative", "key": "connections-closed"},
        "datagrams-received": {"result_type": "relative", "key": "datagrams-received"},
        "datagrams-delivered": {"result_type": "relative", "key": "datagrams-delivered"},
        # Queue Metrics
        "queue-counters-queued-packets": {"result_type": "relative", "key": "q-queued-pkts"},
        "queue-counters-trans-packets": {"result_type": "relative", "key": "q-trans-pkts"},
        "queue-counters-total-drop-packets": {"result_type": "relative", "key": "q-drop-pkts"},
        # Interface Status
        "oper-status": {"result_type": "absolute", "key": "if-status"},
        # Traffic Metrics
        "input-bytes": {"result_type": "per_second", "key": "input-bytes"},
        "output-bytes": {"result_type": "per_second", "key": "output-bytes"},
        # Input Error Metrics
        "input-errors": {"result_type": "relative", "key": "input-errors"},
        "input-drops": {"result_type": "relative", "key": "input-drops"},
        "framing-errors": {"result_type": "relative", "key": "framing-errors"},
        "input-discards": {"result_type": "relative", "key": "input-discards"},
        # Output Error Metrics
        "output-errors": {"result_type": "relative", "key": "output-errors"},
        "output-collisions": {"result_type": "relative", "key": "output-collisions"},
        "output-drops": {"result_type": "relative", "key": "output-drops"},
        # Security Flow Statistics
        "flow-input-policy-bytes": {"result_type": "relative", "key": "flow-input"},
        "flow-output-policy-bytes": {"result_type": "relative", "key": "flow-output"},
    }

    FACTS_TO_COLLECT = {
        "fqdn": "FQDN",
        "hostname": "Hostname",
        "model": "Model",
        "personality": "Personality",
        "serialnumber": "Serial number",
        "switch_style": "Switch style",
        "version": "Version",
        "virtual": "Virtual",
    }

    def initialize(self, **kwargs):
        self._executions = 0
        self._metric_cache: Dict[str, Measurement] = {}

    @exception_logger
    def query(self, **kwargs) -> None:

        self.start_time = time.time()

        self.process_config()
        frequency = self.config.get("poll_frequency", 1)
        if self._executions % frequency == 0:
            self.process_device()
            logger.debug(f"Juniper Execution took {time.time() - self.start_time:.2f} seconds")
        else:
            self.logger.info(f"Not running. {frequency - (self._executions % frequency)} minutes remaining for next run")

        self._executions += 1

    @exception_logger
    def process_config(self) -> None:
        self.check_config()

        self.group = self.topology_builder.create_group(f'Juniper {self.config["group"]}', f'Juniper {self.config["group"]}')

        logger.setLevel(logging.INFO)
        if self.config["debug"]:
            logger.setLevel(logging.DEBUG)

        # Disable INFO logging for ncclient, too much information
        [logging.getLogger(name).setLevel("ERROR") for name in logging.root.manager.loggerDict if "ncc" in name]

    @exception_logger
    def process_device(self) -> None:

        # We support password and ssh private keys, here we figure if the user selected one or the other
        password = None
        ssh_private_key_file = None

        if "auth_ssh_key_file" in self.config and self.config["auth_ssh_key_file"]:
            ssh_private_key_file = self.config["auth_ssh_key_file"]
        else:
            password = self.config["auth_password"]

        self.device = Device(
            host=self.config["device_address"],
            user=self.config["auth_username"],
            password=password,
            ssh_private_key_file=ssh_private_key_file,
            port=self.config["device_port"],
        )

        self.regex_ignore = []
        if "interfaces_to_ignore" in self.config:
            self.regex_ignore = [re.compile(regex) for regex in self.config["interfaces_to_ignore"].split(",")]

        self.regex_include = []
        if "interfaces_to_include" in self.config and self.config["interfaces_to_include"]:
            self.regex_include = [re.compile(regex) for regex in self.config["interfaces_to_include"].split(",")]

        self.logger.info(f"Ignoring interfaces: {self.regex_ignore}")
        self.logger.info(f"Including interfaces: {self.regex_include}")

        self.node_name = f'Juniper {self.config["device_address"]}:{self.config["device_port"]}'
        if "device_alias" in self.config:
            self.node_name = f'Juniper {self.config["device_alias"]}'

        self.node = self.group.create_device(self.node_name)

        try:

            self.device.open()
            self.device.timeout = COMMANDS_TIMEOUT

            with ThreadPoolExecutor(max_workers=6) as e:
                e.submit(self.collect_facts)
                e.submit(self.collect_metrics_memory)
                e.submit(self.collect_metrics_cpu)
                e.submit(self.collect_metrics_statistics)
                e.submit(self.collect_metrics_interface)

        except Exception as e:
            self.report_metric(self.node, "device-status", 0, "absolute")
            logger.exception(f'Exception during query to device "{self.device}": "{e}"')

        finally:
            num_endpoints = 0
            for ip in self.ips:
                if num_endpoints <= 100 and ip not in [ep[0] for ep in self.node.endpoints]:
                    self.node.add_endpoint(ip)
                    num_endpoints += 1
                else:
                    logger.warning(f'Adding "{ip}" would go over the 100 endpoints limit')

            logger.debug(f"Endpoints: {self.node.endpoints}")

            self.device.close()

    @exception_logger
    def collect_facts(self) -> None:
        if self.device.connected:
            self.report_metric(self.node, "device-status", 100, "absolute")

        # Loop through all properties of the device and report them
        for key, value in self.device.facts.items():
            if key in self.FACTS_TO_COLLECT.keys():
                self.node.report_property(self.FACTS_TO_COLLECT[key], f"{value}")
        logger.debug(f"collect_facts took {time.time() - self.start_time:.2f} seconds")

    @exception_logger
    def collect_metrics_memory(self):
        for statistics in self.device.rpc.get_system_memory_information(normalize=True, local=True).iter("system-memory-information"):
            self.report_children(statistics, children=["system-memory-summary-information"])
        logger.debug(f"collect_metrics_memory took {time.time() - self.start_time:.2f} seconds")

    @exception_logger
    def collect_metrics_cpu(self):
        stats = self.device.cli("show system processes brief")
        load_averages_re = re.compile(r"load averages:\s+([0-9]+\.[0-9]+),")
        load_value = re.search(load_averages_re, stats).group(1).strip()
        self.report_metric(self.node, "load-avg", float(load_value), "absolute")
        logger.debug(f"collect_metrics_cpu took {time.time() - self.start_time:.2f} seconds")

    @exception_logger
    def collect_metrics_statistics(self):
        self.report_children(self.device.rpc.get_statistics_information(normalize=True, local=True), children=["tcp", "udp"])
        logger.debug(f"collect_metrics_statistics took {time.time() - self.start_time:.2f} seconds")

    @exception_logger
    def collect_metrics_interface(self):
        for physical_interface in self.device.rpc.get_interface_information(normalize=True, local=True, statistics=True, detail=True):

            interface_name = physical_interface.findtext("name").strip()
            status = physical_interface.findtext("oper-status").upper().strip()

            if any(regex.match(interface_name) for regex in self.regex_ignore):
                if not any(regex.match(interface_name) for regex in self.regex_include):
                    logger.debug(f"{self.node_name} - Interface {interface_name} matched the regex and will be ignored")
                    continue

            logger.debug(f"{self.node_name} - Obtaining statistics for physical interface {interface_name}")

            # Physical interfaces do not have an IP Address
            if not self.config.get("ignore_when_no_ip", False):

                # Traffic and Error Measures
                self.report_children(
                    physical_interface, children=["traffic-statistics", "input-error-list", "output-error-list"], dimensions={"Interface": interface_name},
                )

                # Queue Measures
                for queues in physical_interface.iterfind("queue-counters"):
                    for queue in queues.iterfind("queue"):
                        queue_name = f'Queue ({interface_name}) - {queue.findtext("queue-number")} ({queue.findtext("forwarding-class-name")})'
                        self.report_children(queue, dimensions={"Queue": queue_name})

                self.report_metric(self.node, "if-status", status, "state", dimensions={"Interface": interface_name})

            # Logical Interface Measures
            for logical_interface in physical_interface.iterfind("logical-interface"):

                logical_interface_name = logical_interface.findtext("name").strip()

                if any(regex.match(logical_interface_name) for regex in self.regex_ignore):
                    if not any(regex.match(logical_interface_name) for regex in self.regex_include):
                        logger.debug(f"{self.node_name} - Interface {logical_interface_name} matched the regex and will be ignored")
                        continue

                logger.debug(f"{self.node_name} - Obtaining statistics for logical interface {interface_name}")
                monitor = True
                if self.config.get("ignore_when_no_ip", False):
                    monitor = False

                # Device IP Addresses Collection
                for address_family in logical_interface.iterfind("address-family"):
                    for address in address_family.iterfind("interface-address"):
                        self.ips.add(address.findtext("ifa-local").strip())
                        monitor = True

                if monitor:
                    self.report_children(
                        logical_interface,
                        children=["traffic-statistics", "security-input-flow-statistics", "security-output-flow-statistics"],
                        dimensions={"Interface": logical_interface_name},
                    )

                    self.report_metric(self.node, "if-status", status, "state", dimensions={"Interface": logical_interface_name})

        logger.debug(f"collect_metrics_interface took {time.time() - self.start_time:.2f} seconds")

    def report_children(self, element: ElementTree, children: List[str] = None, dimensions: Dict[str, str] = None) -> None:
        """
        Loops through the children of an etree.Element and report the metrics under them.
        If children is None, reports the metrics immediately inside the etree.Element

        Example XML:

        <system-memory-information>
            <system-memory-summary-information>
                <system-memory-total>1031572</system-memory-total>
                <system-memory-total-percent>100%</system-memory-total-percent>
                <system-memory-reserved>18688</system-memory-reserved>
            </system-memory-summary-information>
        </system-memory-information>

        Element is <system-memory-information>
        There is only one child, <system-memory-summary-information>
        The method will match metrics present in self.metrics_to_collect and under the child

        :param element: The XML ElementTree where the children to report are
        :type element: lxml.etree.ElementTree
        :param children: List of XML tags, the children to loop through
        :type children: list
        :param dimensions: Dynatrace dimensions dict
        :type dimensions: dict
        :return: None
        """

        statistics = [element]
        if children is not None:
            statistics = element.iter(children)

        for statistic in statistics:
            for metric in statistic.iter(self.metrics_to_collect.keys()):
                self.report_metric(
                    self.node,
                    self.metrics_to_collect[metric.tag]["key"],
                    self.remove_chars(metric.text),
                    self.metrics_to_collect[metric.tag]["result_type"],
                    dimensions=dimensions,
                )

    def check_config(self) -> None:
        error_message = ""
        missing_fields = []

        if "auth_username" not in self.config:
            missing_fields.append("Username")

        if "device_address" not in self.config:
            missing_fields.append("Address")

        if "device_port" not in self.config:
            missing_fields.append("Port")

        if "auth_password" not in self.config and "auth_ssh_key_file" not in self.config:
            missing_fields.append("Password or SSH Key File")

        if missing_fields:
            error_message += f'{", ".join(missing_fields)} {"is" if len(missing_fields) == 1 else "are"} mandatory'

        if error_message:
            raise ConfigException(error_message)

    def report_metric(self, node: Node, key: str, value, measurement_type: str, dimensions: Dict[str, str] = None) -> None:

        if measurement_type == "absolute":
            logger.debug(f"Reporting metric: {key}: {value} ({measurement_type}), dimensions: {dimensions}")
            self.node.absolute(key, value, dimensions=dimensions)

        elif measurement_type == "state":
            logger.debug(f"Reporting metric: {key}: {str(value).upper()} ({measurement_type}), dimensions: {dimensions}")
            self.node.state_metric(key, str(value).upper(), dimensions=dimensions)

        # We can't use relative and per_second methods because values are flushed after every query method execution
        # Here we calculate the relative and per_second values ourselves
        dict_key = f"{key}_{dimensions}"
        old_measurement = self._metric_cache.get(dict_key)
        new_measurement = Measurement(value)

        if old_measurement is not None:

            self.logger.debug(f"dict_key: {dict_key}, new_measurement: {new_measurement}, old_measurement: {old_measurement}")

            if measurement_type == "relative":
                relative_value = max(float(value) - float(old_measurement.value), 0)
                self.logger.debug(f"Reporting metric: {key}: {relative_value} ({measurement_type}), dimensions: {dimensions}")
                node.absolute(key, relative_value, dimensions=dimensions)

            elif measurement_type == "per_second":
                relative_value = max(float(value) - float(old_measurement.value), 0)
                time_elapsed = max(time.time() - old_measurement.timestamp, 1)
                per_second_value = relative_value / time_elapsed
                self.logger.debug(f"Per second calculation: {relative_value}/{time_elapsed} = {per_second_value:.2f}")
                self.logger.debug(f"Reporting metric: {key}: {per_second_value} ({measurement_type}), dimensions: {dimensions}")
                node.absolute(key, per_second_value, dimensions=dimensions)

        self._metric_cache[dict_key] = new_measurement

    @staticmethod
    def remove_chars(s: str) -> str:
        return "".join([c for c in s if c.isdigit()])


def main():
    log = logging.getLogger(__name__)
    log.setLevel(logging.DEBUG)
    st = logging.StreamHandler()
    fmt = logging.Formatter("%(asctime)s - %(levelname)s - %(name)s - %(thread)d - %(filename)s:%(lineno)d - %(message)s")
    st.setFormatter(fmt)
    log.addHandler(st)

    class DummyDevice:
        def absolute(self, *args, **kwargs):
            pass

        def relative(self, *args, **kwargs):
            pass

        def per_second(self, *args, **kwargs):
            pass

    d = DummyDevice()
    c = CustomJuniperPluginRemote(topology_builder={}, results_builder={})
    c._metric_cache = {}
    c.logger = log
    c.report_metric(d, "bytes_in", 100, "relative", dimensions={"Interface": "if1"})
    c.report_metric(d, "bytes_in", 200, "relative", dimensions={"Interface": "if2"})

    c.report_metric(d, "bytes_in", 200, "relative", dimensions={"Interface": "if1"})
    c.report_metric(d, "bytes_in", 300, "relative", dimensions={"Interface": "if2"})

    c.report_metric(d, "bytes_in", 250, "relative", dimensions={"Interface": "if1"})
    c.report_metric(d, "bytes_in", 350, "relative", dimensions={"Interface": "if2"})

    c.report_metric(d, "bytes_in", "275", "relative", dimensions={"Interface": "if1"})
    c.report_metric(d, "bytes_in", "375", "relative", dimensions={"Interface": "if2"})

    c.report_metric(d, "bytes_out", 100, "per_second", dimensions={"Interface": "if1"})
    c.report_metric(d, "bytes_out", 200, "per_second", dimensions={"Interface": "if2"})
    time.sleep(5)
    c.report_metric(d, "bytes_out", 200, "per_second", dimensions={"Interface": "if1"})
    c.report_metric(d, "bytes_out", 300, "per_second", dimensions={"Interface": "if2"})
    time.sleep(1)
    c.report_metric(d, "bytes_out", 250, "per_second", dimensions={"Interface": "if1"})
    c.report_metric(d, "bytes_out", 350, "per_second", dimensions={"Interface": "if2"})


if __name__ == "__main__":
    main()
